@@ -22,34 +22,9 @@
 #include <assert.h>
 
 #include "uv.h"
-#include "../uv-common.h"
 #include "internal.h"
-
-
-void uv_stream_init(uv_loop_t* loop, uv_stream_t* handle) {
-  handle->write_queue_size = 0;
-  handle->loop = loop;
-  handle->flags = 0;
-
-  loop->counters.handle_init++;
-  loop->counters.stream_init++;
-
-  uv_ref(loop);
-}
-
-
-void uv_connection_init(uv_stream_t* handle) {
-  handle->flags |= UV_HANDLE_CONNECTION;
-  handle->write_reqs_pending = 0;
-
-  uv_req_init(handle->loop, (uv_req_t*) &(handle->read_req));
-  handle->read_req.event_handle = NULL;
-  handle->read_req.wait_handle = INVALID_HANDLE_VALUE;
-  handle->read_req.type = UV_READ;
-  handle->read_req.data = handle;
-
-  handle->shutdown_req = NULL;
-}
+#include "handle-inl.h"
+#include "req-inl.h"
 
 
 int uv_listen(uv_stream_t* stream, int backlog, uv_connection_cb cb) {
@@ -80,6 +55,16 @@ int uv_accept(uv_stream_t* server, uv_stream_t* client) {
 
 int uv_read_start(uv_stream_t* handle, uv_alloc_cb alloc_cb,
     uv_read_cb read_cb) {
+  if (handle->flags & UV_HANDLE_READING) {
+    uv__set_artificial_error(handle->loop, UV_EALREADY);
+    return -1;
+  }
+
+  if (!(handle->flags & UV_HANDLE_READABLE)) {
+    uv__set_artificial_error(handle->loop, UV_ENOTCONN);
+    return -1;
+  }
+
   switch (handle->type) {
     case UV_TCP:
       return uv_tcp_read_start((uv_tcp_t*)handle, alloc_cb, read_cb);
@@ -96,6 +81,16 @@ int uv_read_start(uv_stream_t* handle, uv_alloc_cb alloc_cb,
 
 int uv_read2_start(uv_stream_t* handle, uv_alloc_cb alloc_cb,
     uv_read2_cb read_cb) {
+  if (handle->flags & UV_HANDLE_READING) {
+    uv__set_artificial_error(handle->loop, UV_EALREADY);
+    return -1;
+  }
+
+  if (!(handle->flags & UV_HANDLE_READABLE)) {
+    uv__set_artificial_error(handle->loop, UV_ENOTCONN);
+    return -1;
+  }
+
   switch (handle->type) {
     case UV_NAMED_PIPE:
       return uv_pipe_read2_start((uv_pipe_t*)handle, alloc_cb, read_cb);
@@ -107,10 +102,14 @@ int uv_read2_start(uv_stream_t* handle, uv_alloc_cb alloc_cb,
 
 
 int uv_read_stop(uv_stream_t* handle) {
+  if (!(handle->flags & UV_HANDLE_READING))
+    return 0;
+
   if (handle->type == UV_TTY) {
     return uv_tty_read_stop((uv_tty_t*) handle);
   } else {
     handle->flags &= ~UV_HANDLE_READING;
+    DECREASE_ACTIVE_COUNT(handle->loop, handle);
     return 0;
   }
 }
@@ -120,6 +119,11 @@ int uv_write(uv_write_t* req, uv_stream_t* handle, uv_buf_t bufs[], int bufcnt,
     uv_write_cb cb) {
   uv_loop_t* loop = handle->loop;
 
+  if (!(handle->flags & UV_HANDLE_WRITABLE)) {
+    uv__set_artificial_error(loop, UV_EPIPE);
+    return -1;
+  }
+
   switch (handle->type) {
     case UV_TCP:
       return uv_tcp_write(loop, req, (uv_tcp_t*) handle, bufs, bufcnt, cb);
@@ -139,6 +143,11 @@ int uv_write2(uv_write_t* req, uv_stream_t* handle, uv_buf_t bufs[], int bufcnt,
     uv_stream_t* send_handle, uv_write_cb cb) {
   uv_loop_t* loop = handle->loop;
 
+  if (!(handle->flags & UV_HANDLE_WRITABLE)) {
+    uv__set_artificial_error(loop, UV_EPIPE);
+    return -1;
+  }
+
   switch (handle->type) {
     case UV_NAMED_PIPE:
       return uv_pipe_write2(loop, req, (uv_pipe_t*) handle, bufs, bufcnt, send_handle, cb);
@@ -153,13 +162,13 @@ int uv_write2(uv_write_t* req, uv_stream_t* handle, uv_buf_t bufs[], int bufcnt,
 int uv_shutdown(uv_shutdown_t* req, uv_stream_t* handle, uv_shutdown_cb cb) {
   uv_loop_t* loop = handle->loop;
 
-  if (!(handle->flags & UV_HANDLE_CONNECTION)) {
-    uv__set_sys_error(loop, WSAEINVAL);
+ if (!(handle->flags & UV_HANDLE_WRITABLE)) {
+    uv__set_artificial_error(loop, UV_EPIPE);
     return -1;
   }
 
-  if (handle->flags & UV_HANDLE_SHUTTING) {
-    uv__set_sys_error(loop, WSAESHUTDOWN);
+  if (!(handle->flags & UV_HANDLE_WRITABLE)) {
+    uv__set_artificial_error(loop, UV_EPIPE);
     return -1;
   }
 
@@ -168,10 +177,10 @@ int uv_shutdown(uv_shutdown_t* req, uv_stream_t* handle, uv_shutdown_cb cb) {
   req->handle = handle;
   req->cb = cb;
 
-  handle->flags |= UV_HANDLE_SHUTTING;
+  handle->flags &= ~UV_HANDLE_WRITABLE;
   handle->shutdown_req = req;
   handle->reqs_pending++;
-  uv_ref(loop);
+  REGISTER_HANDLE_REQ(loop, handle, req);
 
   uv_want_endgame(loop, (uv_handle_t*)handle);
 
@@ -179,23 +188,11 @@ int uv_shutdown(uv_shutdown_t* req, uv_stream_t* handle, uv_shutdown_cb cb) {
 }
 
 
-size_t uv_count_bufs(uv_buf_t bufs[], int count) {
-  size_t bytes = 0;
-  int i;
-
-  for (i = 0; i < count; i++) {
-    bytes += (size_t)bufs[i].len;
-  }
-
-  return bytes;
+int uv_is_readable(const uv_stream_t* handle) {
+  return !!(handle->flags & UV_HANDLE_READABLE);
 }
 
 
-int uv_is_readable(uv_stream_t* handle) {
-  return !(handle->flags & UV_HANDLE_EOF);
-}
-
-
-int uv_is_writable(uv_stream_t* handle) {
-  return !(handle->flags & UV_HANDLE_SHUTTING);
+int uv_is_writable(const uv_stream_t* handle) {
+  return !!(handle->flags & UV_HANDLE_WRITABLE);
 }
