@@ -36,11 +36,11 @@ exports.start = function(argv, stdin, stdout) {
   }
 
   // Setup input/output streams
-  stdin = stdin || process.openStdin();
+  stdin = stdin || process.stdin;
   stdout = stdout || process.stdout;
 
   var args = ['--debug-brk'].concat(argv),
-      interface = new Interface(stdin, stdout, args);
+      interface_ = new Interface(stdin, stdout, args);
 
   stdin.resume();
 
@@ -49,7 +49,7 @@ exports.start = function(argv, stdin, stdout) {
         'Please report this bug.');
     console.error(e.message);
     console.error(e.stack);
-    if (interface.child) interface.child.kill();
+    if (interface_.child) interface_.child.kill();
     process.exit(1);
   });
 };
@@ -112,8 +112,8 @@ Protocol.prototype.execute = function(d) {
 
       this.state = 'body';
 
-      if (Buffer.byteLength(res.raw, 'utf8') - this.bodyStartByteIndex
-          < this.contentLength) {
+      var len = Buffer.byteLength(res.raw, 'utf8');
+      if (len - this.bodyStartByteIndex < this.contentLength) {
         break;
       }
       // pass thru
@@ -125,16 +125,16 @@ Protocol.prototype.execute = function(d) {
         buf.write(res.raw, 0, resRawByteLength, 'utf8');
         res.body =
             buf.slice(this.bodyStartByteIndex,
-                      this.bodyStartByteIndex
-                      + this.contentLength).toString('utf8');
+                      this.bodyStartByteIndex +
+                      this.contentLength).toString('utf8');
         // JSON parse body?
         res.body = res.body.length ? JSON.parse(res.body) : {};
 
         // Done!
         this.onResponse(res);
 
-        this._newRes(buf.slice(this.bodyStartByteIndex
-                               + this.contentLength).toString('utf8'));
+        this._newRes(buf.slice(this.bodyStartByteIndex +
+                               this.contentLength).toString('utf8'));
       }
       break;
 
@@ -149,8 +149,8 @@ Protocol.prototype.serialize = function(req) {
   req.type = 'request';
   req.seq = this.reqSeq++;
   var json = JSON.stringify(req);
-  return 'Content-Length: ' + Buffer.byteLength(json,'utf8') + '\r\n\r\n'
-      + json;
+  return 'Content-Length: ' + Buffer.byteLength(json, 'utf8') +
+         '\r\n\r\n' + json;
 };
 
 
@@ -236,6 +236,10 @@ Client.prototype._onResponse = function(res) {
     this.emit('break', res.body);
     handled = true;
 
+  } else if (res.body && res.body.event == 'exception') {
+    this.emit('exception', res.body);
+    handled = true;
+
   } else if (res.body && res.body.event == 'afterCompile') {
     this._addHandle(res.body.body.script);
     handled = true;
@@ -409,6 +413,16 @@ Client.prototype.reqBacktrace = function(cb) {
 };
 
 
+// reqSetExceptionBreak(type, cb)
+// TODO: from, to, bottom
+Client.prototype.reqSetExceptionBreak = function(type, cb) {
+  this.req({
+    command: 'setexceptionbreak',
+    arguments: { type: type, enabled: true }
+  }, cb);
+};
+
+
 // Returns an array of objects like this:
 //
 //   { handle: 11,
@@ -495,7 +509,7 @@ Client.prototype.mirrorObject = function(handle, depth, cb) {
 
   var val;
 
-  if (handle.type == 'object') {
+  if (handle.type === 'object') {
     // The handle looks something like this:
     // { handle: 8,
     //   type: 'object',
@@ -527,6 +541,8 @@ Client.prototype.mirrorObject = function(handle, depth, cb) {
 
       if (handle.className == 'Array') {
         mirror = [];
+      } else if (handle.className == 'Date') {
+        mirror = new Date(handle.value);
       } else {
         mirror = {};
       }
@@ -664,6 +680,7 @@ var commands = [
     'kill',
     'list',
     'scripts',
+    'breakOnException',
     'breakpoints',
     'version'
   ]
@@ -675,15 +692,15 @@ var helpMessage = 'Commands: ' + commands.map(function(group) {
 }).join(',\n');
 
 
-function SourceUnderline(sourceText, position, tty) {
+function SourceUnderline(sourceText, position, repl) {
   if (!sourceText) return '';
 
   var head = sourceText.slice(0, position),
       tail = sourceText.slice(position);
 
   // Colourize char if stdout supports colours
-  if (tty && !repl.disableColors) {
-    tail = tail.replace(/(.+?)([^\w]|$)/, '\033[32m$1\033[39m$2');
+  if (repl.useColors) {
+    tail = tail.replace(/(.+?)([^\w]|$)/, '\u001b[32m$1\u001b[39m$2');
   }
 
   // Return source line with coloured char at `position`
@@ -695,7 +712,7 @@ function SourceUnderline(sourceText, position, tty) {
 
 
 function SourceInfo(body) {
-  var result = 'break in ';
+  var result = body.exception ? 'exception in ' : 'break in ';
 
   if (body.script) {
     if (body.script.name) {
@@ -715,6 +732,8 @@ function SourceInfo(body) {
   result += ':';
   result += body.sourceLine + 1;
 
+  if (body.exception) result += '\n' + body.exception.text;
+
   return result;
 }
 
@@ -728,26 +747,45 @@ function Interface(stdin, stdout, args) {
   this.stdout = stdout;
   this.args = args;
 
-  var streams = {
-    stdin: stdin,
-    stdout: stdout
-  };
-
   // Two eval modes are available: controlEval and debugEval
   // But controlEval is used by default
-  this.repl = new repl.REPLServer('debug> ', streams,
-                                  this.controlEval.bind(this), false, true);
+  var opts = {
+    prompt: 'debug> ',
+    input: this.stdin,
+    output: this.stdout,
+    eval: this.controlEval.bind(this),
+    useGlobal: false,
+    ignoreUndefined: true
+  };
+  if (parseInt(process.env['NODE_NO_READLINE'], 10)) {
+    opts.terminal = false;
+  } else if (parseInt(process.env['NODE_FORCE_READLINE'], 10)) {
+    opts.terminal = true;
 
-  // Kill child process when repl closed or main process is dead
-  this.repl.rli.addListener('close', function() {
-    self.killed = true;
-    self.killChild();
-  });
+    // Emulate Ctrl+C if we're emulating terminal
+    if (!this.stdout.isTTY) {
+      process.on('SIGINT', function() {
+        self.repl.rli.emit('SIGINT');
+      });
+    }
+  }
+  if (parseInt(process.env['NODE_DISABLE_COLORS'], 10)) {
+    opts.useColors = false;
+  }
+  this.repl = repl.start(opts);
 
-  process.on('exit', function() {
-    self.killChild();
+  // Do not print useless warning
+  repl._builtinLibs.splice(repl._builtinLibs.indexOf('repl'), 1);
+
+  // Kill child process when main process dies
+  this.repl.on('exit', function() {
+    process.exit(0);
   });
 
+  // Handle all possible exits
+  process.on('exit', this.killChild.bind(this));
+  process.once('SIGTERM', process.exit.bind(process, 0));
+  process.once('SIGHUP', process.exit.bind(process, 0));
 
   var proto = Interface.prototype,
       ignored = ['pause', 'resume', 'exitRepl', 'handleBreak',
@@ -762,7 +800,8 @@ function Interface(stdin, stdout, args) {
         'out': 'o',
         'backtrace': 'bt',
         'setBreakpoint': 'sb',
-        'clearBreakpoint': 'cb'
+        'clearBreakpoint': 'cb',
+        'pause_': 'pause'
       };
 
   function defineProperty(key, protoKey) {
@@ -867,7 +906,7 @@ Interface.prototype.childPrint = function(text) {
   }).map(function(chunk) {
     return '< ' + chunk;
   }).join('\n'));
-  this.repl.displayPrompt();
+  this.repl.displayPrompt(true);
 };
 
 // Errors formatting
@@ -922,7 +961,7 @@ Interface.prototype.controlEval = function(code, context, filename, callback) {
   try {
     // Repeat last command if empty line are going to be evaluated
     if (this.repl.rli.history && this.repl.rli.history.length > 0) {
-      if (code === '(undefined\n)') {
+      if (code === '(\n)') {
         code = '(' + this.repl.rli.history[0] + '\n)';
       }
     }
@@ -1091,8 +1130,9 @@ Interface.prototype.list = function(delta) {
 
       var current = lineno == 1 + client.currentSourceLine,
           breakpoint = client.breakpoints.some(function(bp) {
-            return bp.script === client.currentScript &&
-                   bp.line == lineno;
+            return (bp.scriptReq === client.currentScript ||
+                    bp.script === client.currentScript) &&
+                    bp.line == lineno;
           });
 
       if (lineno == 1) {
@@ -1109,7 +1149,7 @@ Interface.prototype.list = function(delta) {
       if (current) {
         line = SourceUnderline(lines[i],
                                client.currentSourceColumn,
-                               self.stdout.isTTY);
+                               self.repl);
       } else {
         line = lines[i];
       }
@@ -1289,6 +1329,19 @@ Interface.prototype.watchers = function() {
   }
 };
 
+// Break on exception
+Interface.prototype.breakOnException = function breakOnException() {
+  if (!this.requireConnection()) return;
+
+  var self = this;
+
+  // Break on exceptions
+  this.pause();
+  this.client.reqSetExceptionBreak('all', function(err, res) {
+    self.resume();
+  });
+};
+
 // Add breakpoint
 Interface.prototype.setBreakpoint = function(script, line,
                                              condition, silent) {
@@ -1304,6 +1357,12 @@ Interface.prototype.setBreakpoint = function(script, line,
     line = this.client.currentSourceLine + 1;
   }
 
+  // setBreakpoint(line-number) should insert breakpoint in current script
+  if (line === undefined && typeof script === 'number') {
+    line = script;
+    script = this.client.currentScript;
+  }
+
   if (/\(\)$/.test(script)) {
     // setBreakpoint('functionname()');
     var req = {
@@ -1316,7 +1375,9 @@ Interface.prototype.setBreakpoint = function(script, line,
     if (script != +script && !this.client.scripts[script]) {
       var scripts = this.client.scripts;
       Object.keys(scripts).forEach(function(id) {
-        if (scripts[id] && scripts[id].name.indexOf(script) !== -1) {
+        if (scripts[id] &&
+            scripts[id].name &&
+            scripts[id].name.indexOf(script) !== -1) {
           if (scriptId) {
             ambiguous = true;
           }
@@ -1327,16 +1388,28 @@ Interface.prototype.setBreakpoint = function(script, line,
       scriptId = script;
     }
 
-    if (!scriptId) return this.error('Script : ' + script + ' not found');
     if (ambiguous) return this.error('Script name is ambiguous');
     if (line <= 0) return this.error('Line should be a positive value');
 
-    var req = {
-      type: 'scriptId',
-      target: scriptId,
-      line: line - 1,
-      condition: condition
-    };
+    var req;
+    if (scriptId) {
+      req = {
+        type: 'scriptId',
+        target: scriptId,
+        line: line - 1,
+        condition: condition
+      };
+    } else {
+      this.print('Warning: script \'' + script + '\' was not loaded yet.');
+      var escapedPath = script.replace(/([/\\.?*()^${}|[\]])/g, '\\$1');
+      var scriptPathRegex = '^(.*[\\/\\\\])?' + escapedPath + '$';
+      req = {
+        type: 'scriptRegExp',
+        target: scriptPathRegex,
+        line: line - 1,
+        condition: condition
+      };
+    }
   }
 
   self.pause();
@@ -1353,20 +1426,18 @@ Interface.prototype.setBreakpoint = function(script, line,
       // Try load scriptId and line from response
       if (!scriptId) {
         scriptId = res.script_id;
-        line = res.line;
-      }
-
-      // If we finally have one - remember this breakpoint
-      if (scriptId) {
-        self.client.breakpoints.push({
-          id: res.breakpoint,
-          scriptId: scriptId,
-          script: (self.client.scripts[scriptId] || {}).name,
-          line: line,
-          condition: condition
-        });
+        line = res.line + 1;
       }
 
+      // Remember this breakpoint even if scriptId is not resolved yet
+      self.client.breakpoints.push({
+        id: res.breakpoint,
+        scriptId: scriptId,
+        script: (self.client.scripts[scriptId] || {}).name,
+        line: line,
+        condition: condition,
+        scriptReq: script
+      });
     }
     self.resume();
   });
@@ -1381,7 +1452,9 @@ Interface.prototype.clearBreakpoint = function(script, line) {
       index;
 
   this.client.breakpoints.some(function(bp, i) {
-    if (bp.scriptId === script || bp.script.indexOf(script) !== -1) {
+    if (bp.scriptId === script ||
+        bp.scriptReq === script ||
+        (bp.script && bp.script.indexOf(script) !== -1)) {
       if (index !== undefined) {
         ambiguous = true;
       }
@@ -1434,6 +1507,24 @@ Interface.prototype.breakpoints = function() {
 };
 
 
+// Pause child process
+Interface.prototype.pause_ = function() {
+  if (!this.requireConnection()) return;
+
+  var self = this,
+      cmd = 'process._debugPause();';
+
+  this.pause();
+  this.client.reqFrameEval(cmd, NO_FRAME, function(err, res) {
+    if (err) {
+      self.error(err);
+    } else {
+      self.resume();
+    }
+  });
+};
+
+
 // Kill child process
 Interface.prototype.kill = function() {
   if (!this.child) return;
@@ -1450,7 +1541,7 @@ Interface.prototype.repl = function() {
   self.print('Press Ctrl + C to leave debug repl');
 
   // Don't display any default messages
-  var listeners = this.repl.rli.listeners('SIGINT');
+  var listeners = this.repl.rli.listeners('SIGINT').slice(0);
   this.repl.rli.removeAllListeners('SIGINT');
 
   // Exit debug repl on Ctrl + C
@@ -1525,15 +1616,17 @@ Interface.prototype.trySpawn = function(cb) {
   var self = this,
       breakpoints = this.breakpoints || [],
       port = exports.port,
-      host = 'localhost';
+      host = 'localhost',
+      childArgs = this.args;
 
   this.killChild();
 
-  // Connecting to remote debugger
-  // `node debug localhost:5858`
   if (this.args.length === 2) {
     var match = this.args[1].match(/^([^:]+):(\d+)$/);
+
     if (match) {
+      // Connecting to remote debugger
+      // `node debug localhost:5858`
       host = match[1];
       port = parseInt(match[2], 10);
       this.child = {
@@ -1551,11 +1644,19 @@ Interface.prototype.trySpawn = function(cb) {
         }
       };
       process._debugProcess(parseInt(this.args[2], 10));
+    } else {
+      var match = this.args[1].match(/^--port=(\d+)$/);
+      if (match) {
+        // Start debugger on custom port
+        // `node debug --port=5858 app.js`
+        port = parseInt(match[1], 10);
+        childArgs = ['--debug-brk=' + port].concat(this.args.slice(2));
+      }
     }
   }
 
   if (!this.child) {
-    this.child = spawn(process.execPath, this.args);
+    this.child = spawn(process.execPath, childArgs);
 
     this.child.stdout.on('data', this.childPrint.bind(this));
     this.child.stderr.on('data', this.childPrint.bind(this));
@@ -1571,13 +1672,10 @@ Interface.prototype.trySpawn = function(cb) {
 
     // Restore breakpoints
     breakpoints.forEach(function(bp) {
-      self.setBreakpoint(bp.scriptId, bp.line, bp.condition, true);
+      self.print('Restoring breakpoint ' + bp.scriptReq + ':' + bp.line);
+      self.setBreakpoint(bp.scriptReq, bp.line, bp.condition, true);
     });
 
-    if (cb) cb();
-
-    self.resume();
-
     client.on('close', function() {
       self.pause();
       self.print('program terminated');
@@ -1585,6 +1683,9 @@ Interface.prototype.trySpawn = function(cb) {
       self.client = null;
       self.killChild();
     });
+
+    if (cb) cb();
+    self.resume();
   });
 
   client.on('unhandledResponse', function(res) {
@@ -1597,6 +1698,10 @@ Interface.prototype.trySpawn = function(cb) {
     self.handleBreak(res.body);
   });
 
+  client.on('exception', function(res) {
+    self.handleBreak(res.body);
+  });
+
   client.on('error', connectError);
   function connectError() {
     // If it's failed to connect 4 times then don't catch the next error
