@@ -344,9 +344,7 @@ IncomingMessage.prototype._read = function(n) {
   // We actually do almost nothing here, because the parserOnBody
   // function fills up our internal buffer directly.  However, we
   // do need to unpause the underlying socket so that it flows.
-  if (!this.socket.readable)
-    this.push(null);
-  else
+  if (this.socket.readable)
     readStart(this.socket);
 };
 
@@ -1346,7 +1344,9 @@ function ClientRequest(options, cb) {
   var self = this;
   OutgoingMessage.call(self);
 
-  self.agent = options.agent === undefined ? globalAgent : options.agent;
+  self.agent = options.agent;
+  if (!options.agent && options.agent !== false && !options.createConnection)
+    self.agent = globalAgent;
 
   var defaultPort = options.defaultPort || 80;
 
@@ -2078,7 +2078,7 @@ function connectionListener(socket) {
       // if the user never called req.read(), and didn't pipe() or
       // .resume() or .on('data'), then we call req._dump() so that the
       // bytes will be pulled off the wire.
-      if (!req._consuming)
+      if (!req._consuming && !req._readableState.oldMode)
         req._dump();
 
       res.detachSocket(socket);
