@@ -217,7 +217,7 @@ node_module_struct* get_builtin_module(const char *name);
 #ifdef _WIN32
 # define NODE_MODULE_EXPORT __declspec(dllexport)
 #else
-# define NODE_MODULE_EXPORT /* empty */
+# define NODE_MODULE_EXPORT __attribute__((visibility("default")))
 #endif
 
 #define NODE_MODULE(modname, regfunc)                                 \
