@@ -20,6 +20,8 @@
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 
+#include "util.h"
+
 #ifdef HAVE_DTRACE
 #include "node_dtrace.h"
 #include <string.h>
@@ -66,7 +68,7 @@ using namespace v8;
     return (ThrowException(Exception::Error(String::New("expected " \
       "object for " #obj " to contain string member " #member)))); \
   } \
-  String::Utf8Value _##member(obj->Get(String::New(#member))); \
+  node::Utf8Value _##member(obj->Get(String::New(#member))); \
   if ((*(const char **)valp = *_##member) == NULL) \
     *(const char **)valp = "<unknown>";
 
@@ -241,7 +243,7 @@ Handle<Value> DTRACE_HTTP_SERVER_REQUEST(const Arguments& args) {
       "object for request to contain string member headers"))));
 
   Local<Value> strfwdfor = headers->Get(String::New("x-forwarded-for"));
-  String::Utf8Value fwdfor(strfwdfor);
+  node::Utf8Value fwdfor(strfwdfor);
 
   if (!strfwdfor->IsString() || (req.forwardedFor = *fwdfor) == NULL)
     req.forwardedFor = const_cast<char*>("");
@@ -341,7 +343,7 @@ Handle<Value> DTRACE_HTTP_CLIENT_RESPONSE(const Arguments& args) {
 
 #define NODE_PROBE(name) #name, name, Persistent<FunctionTemplate>()
 
-static int dtrace_gc_start(GCType type, GCCallbackFlags flags) {
+int dtrace_gc_start(GCType type, GCCallbackFlags flags) {
 #ifdef HAVE_SYSTEMTAP
   NODE_GC_START();
 #else
@@ -354,7 +356,7 @@ static int dtrace_gc_start(GCType type, GCCallbackFlags flags) {
   return 0;
 }
 
-static int dtrace_gc_done(GCType type, GCCallbackFlags flags) {
+int dtrace_gc_done(GCType type, GCCallbackFlags flags) {
 #ifdef HAVE_SYSTEMTAP
   NODE_GC_DONE();
 #else
