@@ -249,6 +249,11 @@ try {
   gotError = true;
 }
 
+// GH-7178. Ensure reflexivity of deepEqual with `arguments` objects.
+var args = (function() { return arguments; })();
+a.throws(makeBlock(a.deepEqual, [], args));
+a.throws(makeBlock(a.deepEqual, args, []));
+
 console.log('All OK');
 assert.ok(gotError);
 
