@@ -49,8 +49,8 @@ function runAb(opts, callback) {
   var command = 'ab ' + opts + ' http://127.0.0.1:' + common.PORT + '/';
   exec(command, function(err, stdout, stderr) {
     if (err) {
-      if (stderr.indexOf('ab') >= 0) {
-        console.log('ab not installed? skipping test.\n' + stderr);
+      if (/ab|apr/mi.test(stderr)) {
+        console.log('problem spawning ab - skipping test.\n' + stderr);
         process.reallyExit(0);
       }
       process.exit();
