@@ -41,12 +41,14 @@ for (var i = 255; i >= 0; i--) {
 
 // safe constructor
 var echoServer = net.Server(function(connection) {
+  console.error('SERVER got connection');
   connection.setEncoding('binary');
-  connection.addListener('data', function(chunk) {
-    common.error('recved: ' + JSON.stringify(chunk));
+  connection.on('data', function(chunk) {
+    common.error('SERVER recved: ' + JSON.stringify(chunk));
     connection.write(chunk, 'binary');
   });
-  connection.addListener('end', function() {
+  connection.on('end', function() {
+    console.error('SERVER ending');
     connection.end();
   });
 });
@@ -54,33 +56,46 @@ echoServer.listen(common.PORT);
 
 var recv = '';
 
-echoServer.addListener('listening', function() {
+echoServer.on('listening', function() {
+  console.error('SERVER listening');
   var j = 0;
-  var c = net.createConnection(common.PORT);
+  var c = net.createConnection({
+    port: common.PORT
+  });
 
   c.setEncoding('binary');
-  c.addListener('data', function(chunk) {
-    if (j < 256) {
-      common.error('write ' + j);
+  c.on('data', function(chunk) {
+    console.error('CLIENT data %j', chunk);
+    var n = j + chunk.length;
+    while (j < n && j < 256) {
+      common.error('CLIENT write ' + j);
       c.write(String.fromCharCode(j), 'binary');
       j++;
-    } else {
+    }
+    if (j === 256) {
+      console.error('CLIENT ending');
       c.end();
     }
     recv += chunk;
   });
 
-  c.addListener('connect', function() {
+  c.on('connect', function() {
+    console.error('CLIENT connected, writing');
     c.write(binaryString, 'binary');
   });
 
-  c.addListener('close', function() {
+  c.on('close', function() {
+    console.error('CLIENT closed');
     console.dir(recv);
     echoServer.close();
   });
+
+  c.on('finish', function() {
+    console.error('CLIENT finished');
+  });
 });
 
-process.addListener('exit', function() {
+process.on('exit', function() {
   console.log('recv: ' + JSON.stringify(recv));
 
   assert.equal(2 * 256, recv.length);
