@@ -31,7 +31,6 @@
 import imp
 import optparse
 import os
-from os.path import join, dirname, abspath, basename, isdir, exists
 import platform
 import re
 import signal
@@ -40,9 +39,13 @@ import sys
 import tempfile
 import time
 import threading
-import utils
+
+from os.path import join, dirname, abspath, basename, isdir, exists
+from datetime import datetime
 from Queue import Queue, Empty
 
+sys.path.append(dirname(__file__) + "/../deps/v8/tools");
+import utils
 
 VERBOSE = False
 
@@ -114,9 +117,9 @@ class ProgressIndicator(object):
       self.AboutToRun(case)
       self.lock.release()
       try:
-        start = time.time()
+        start = datetime.now()
         output = case.Run()
-        case.duration = (time.time() - start)
+        case.duration = (datetime.now() - start)
       except IOError, e:
         assert self.terminate
         return
@@ -220,6 +223,41 @@ class DotsProgressIndicator(SimpleProgressIndicator):
       sys.stdout.flush()
 
 
+class TapProgressIndicator(SimpleProgressIndicator):
+
+  def Starting(self):
+    print '1..%i' % len(self.cases)
+    self._done = 0
+
+  def AboutToRun(self, case):
+    pass
+
+  def HasRun(self, output):
+    self._done += 1
+    command = basename(output.command[-1])
+    if output.UnexpectedOutput():
+      print 'not ok %i - %s' % (self._done, command)
+      for l in output.output.stderr.splitlines():
+        print '#' + l
+      for l in output.output.stdout.splitlines():
+        print '#' + l
+    else:
+      print 'ok %i - %s' % (self._done, command)
+
+    duration = output.test.duration
+
+    # total_seconds() was added in 2.7
+    total_seconds = (duration.microseconds +
+      (duration.seconds + duration.days * 24 * 3600) * 10**6) / 10**6
+
+    print '  ---'
+    print '  duration_ms: %d.%d' % (total_seconds, duration.microseconds / 1000)
+    print '  ...'
+
+  def Done(self):
+    pass
+
+
 class CompactProgressIndicator(ProgressIndicator):
 
   def __init__(self, cases, templates):
@@ -310,6 +348,7 @@ PROGRESS_INDICATORS = {
   'verbose': VerboseProgressIndicator,
   'dots': DotsProgressIndicator,
   'color': ColorProgressIndicator,
+  'tap': TapProgressIndicator,
   'mono': MonochromeProgressIndicator
 }
 
@@ -373,10 +412,20 @@ class TestCase(object):
 
   def Run(self):
     self.BeforeRun()
+
     try:
       result = self.RunCommand(self.GetCommand())
     finally:
-      self.AfterRun(result)
+      # Tests can leave the tty in non-blocking mode. If the test runner
+      # tries to print to stdout/stderr after that and the tty buffer is
+      # full, it'll die with a EAGAIN OSError. Ergo, put the tty back in
+      # blocking mode before proceeding.
+      if sys.platform != 'win32':
+        from fcntl import fcntl, F_GETFL, F_SETFL
+        from os import O_NONBLOCK
+        for fd in 0,1,2: fcntl(fd, F_SETFL, ~O_NONBLOCK & fcntl(fd, F_GETFL))
+
+    self.AfterRun(result)
     return result
 
   def Cleanup(self):
@@ -454,7 +503,6 @@ def RunProcess(context, timeout, args, **rest):
   popen_args = args
   prev_error_mode = SEM_INVALID_VALUE;
   if utils.IsWindows():
-    popen_args = '"' + subprocess.list2cmdline(args) + '"'
     if context.suppress_dialogs:
       # Try to change the error mode to avoid dialogs on fatal errors. Don't
       # touch any existing error mode flags by merging the existing error mode.
@@ -631,6 +679,7 @@ class LiteralTestSuite(TestSuite):
       if not name or name.match(test_name):
         full_path = current_path + [test_name]
         test.AddTestsToList(result, full_path, path, context, mode)
+    result.sort(cmp=lambda a, b: cmp(a.GetName(), b.GetName()))
     return result
 
   def GetTestStatus(self, context, sections, defs):
@@ -663,12 +712,23 @@ class Context(object):
 
   def GetVm(self, mode):
     if mode == 'debug':
-      name = 'build/debug/node_g'
+      name = 'out/Debug/node'
     else:
-      name = 'build/default/node'
+      name = 'out/Release/node'
+
+    # Currently GYP does not support output_dir for MSVS.
+    # http://code.google.com/p/gyp/issues/detail?id=40
+    # It will put the builds into Release/node.exe or Debug/node.exe
+    if utils.IsWindows():
+      out_dir = os.path.join(dirname(__file__), "..", "out")
+      if not exists(out_dir):
+        if mode == 'debug':
+          name = os.path.abspath('Debug/node.exe')
+        else:
+          name = os.path.abspath('Release/node.exe')
+      else:
+        name = os.path.abspath(name + '.exe')
 
-    if utils.IsWindows() and not name.endswith('.exe'):
-      name = name + '.exe'
     return name
 
   def GetVmCommand(self, testcase, mode):
@@ -1129,7 +1189,7 @@ def BuildOptions():
   result.add_option("-S", dest="scons_flags", help="Flag to pass through to scons",
       default=[], action="append")
   result.add_option("-p", "--progress",
-      help="The style of progress indicator (verbose, dots, color, mono)",
+      help="The style of progress indicator (verbose, dots, color, mono, tap)",
       choices=PROGRESS_INDICATORS.keys(), default="mono")
   result.add_option("--no-build", help="Don't build requirements",
       default=True, action="store_true")
@@ -1148,6 +1208,8 @@ def BuildOptions():
   result.add_option("--simulator", help="Run tests with architecture simulator",
       default='none')
   result.add_option("--special-command", default=None)
+  result.add_option("--use-http1", help="Pass --use-http1 switch to node",
+      default=False, action="store_true")
   result.add_option("--valgrind", help="Run tests through valgrind",
       default=False, action="store_true")
   result.add_option("--cat", help="Print the source of the tests",
@@ -1262,7 +1324,7 @@ def GetSpecialCommandProcessor(value):
     return ExpandCommand
 
 
-BUILT_IN_TESTS = ['simple', 'pummel', 'message', 'internet']
+BUILT_IN_TESTS = ['simple', 'pummel', 'message', 'internet', 'gc']
 
 
 def GetSuites(test_root):
@@ -1306,10 +1368,18 @@ def Main():
   shell = abspath(options.shell)
   buildspace = dirname(shell)
 
-  context = Context(workspace, buildspace, VERBOSE,
+  processor = GetSpecialCommandProcessor(options.special_command)
+  if options.use_http1:
+    def wrap(processor):
+      return lambda args: processor(args[:1] + ['--use-http1'] + args[1:])
+    processor = wrap(processor)
+
+  context = Context(workspace,
+                    buildspace,
+                    VERBOSE,
                     shell,
                     options.timeout,
-                    GetSpecialCommandProcessor(options.special_command),
+                    processor,
                     options.suppress_dialogs,
                     options.store_unexpected_output)
   # First build the required targets
